-----------------------------------------------------------------------------
--	File	 :	/lua/sim/Projectile.lua
--	Author(s):	Resin Smoker
--	Summary  :	Base Projectile Definition for Projectile Reflections
--	Copyright 2014 All rights reserved.
--	Original Projectile.lua size 465 lines
-----------------------------------------------------------------------------
do--(start of non-destructive hook)

local oldProjectile = Projectile
local utilities = import('/lua/utilities.lua')
local EffectTemplate = import('/lua/EffectTemplates.lua')

--> Triggers all logs to report if flag set true, selfwise use the local flags within each function for specific outputs
local masterDebug = false

local AllowProjTimeOut = true		--> Remove projectiles that have reached the end of their lifetime
local minDamage = 1					--> Minimum damage the projectile will be allowed to do
local maxProjLifeTime = 300			--> The longest amount of time in seconds the projectile will stay alive
local minArtyLifeTime = 20			--> The shortest amount of time in seconds the directfired projectile will stay alive
local minBombLifeTime = 10			--> The shortest amount of time in seconds the bomb projectile will stay alive
local minDirectLifeTime = 2			--> The shortest amount of time in seconds the directfired projectile will stay alive
local maxProjVelocity = 160			--> The absolute highest speed a projectile is allowed to to move
local minArtyVelocity = 8			--> The slowest speed a artillery projectile will travel
local minBombVelocity = 4			--> The slowest speed a bomb projectile will travel
local minDirectVelocity = 8			--> The slowest speed a directfired projectile will travel

Projectile = Class( oldProjectile ) {

	OnCreate = function( self, inWater )
		local myDebug = false
		if masterDebug and myDebug then
			LOG('OnCreate (/hook/lua/sim/Projectile.lua), Projectile ID ', self:GetEntityId() )
		end
		--> Run oldProjectile
		oldProjectile.OnCreate( self, inWater )
	end,

	PassDamageData = function( self, damageData )
		local myDebug = false
		if masterDebug and myDebug then
			LOG('	PassData (/hook/lua/sim/Projectile.lua), Projectile ID ', self:GetEntityId(), '	DamageData.DamageAmount ', repr( damageData.DamageAmount ) )
			LOG('BP ', repr( self:GetBlueprint().BlueprintId ) )
		end
		--> Run oldProjectile, please be aware that PassDamageData is not used by any projectiles altered by reflections and in many cases this can be true for custom projectiles as well
		oldProjectile.PassDamageData( self, damageData )
		--> Save the projectiles starting location for later comparison
		self.DamageData.LauncherLoc = {}
		if self and not self:BeenDestroyed() and self:GetLauncher() then
			self.DamageData.LauncherLoc = self:GetLauncher():GetPosition()
		end
	end,
	
	ValidateProjReflect = function( self, targetEntity )
		local myDebug = false
		if masterDebug and myDebug then
			LOG('	ValidateProjReflect (/hook/lua/sim/Projectile.lua), Projectile ID ', self:GetEntityId() )
		end
		--> Ensure that the launching unit is avalible		
		if self:GetLauncher() then	
			--> Remember the projectiles BP
			if self:SetProjBP() then
				--> Remember the original launcher
				if self:SetLauncherUnit() then
					--> Remember the launchers BP
					if self:SetLauncherBP() then						
						--> Detirmine which weapon fired the projectile
						if self:SetLauncherWeapon() then			
							--> Determine Velocity
							if self:SetProjVelocity() then
								--> Determine Lifetime					
								if self:SetProjLifetime() then																		
									--> Determine the return Vector						
									if self:SetProjVectors( targetEntity ) then										
										--> Set the projectiles global incase there is another impact later with this shield
										self:SetReflected( targetEntity )
										if masterDebug and myDebug then
											LOG('		ValidateProjReflect (/hook/lua/sim/Projectile.lua), Calling for projectile to reflect ','	Projectile ID ', self:GetEntityId() )
										end
										--> Reflect !!!			
										self:ProjReflect( targetEntity )
										if masterDebug and myDebug then
											LOG('		ValidateProjReflect (/hook/lua/sim/Projectile.lua), Returning true to allow this projectile to reflect ','	Projectile ID ', self:GetEntityId() )
										end										
										return true
									end										
								end
							end
						end
					end
				end
			end
		else
			if masterDebug and myDebug then
				LOG('		ValidateProjReflect (/hook/lua/sim/Projectile.lua), Launcher not avalible or missing')
			end						
		end	
		if masterDebug and myDebug then
			LOG('		ValidateProjReflect (/hook/lua/sim/Projectile.lua), Returning false preventing this projectile from reflecting')
		end		
		return false		
	end,
	
	SetProjBP = function( self )
		local myDebug = false	
		if masterDebug and myDebug then
			LOG('	SetProjBP (/hook/lua/sim/Projectile.lua), Projectile ID ', self:GetEntityId() )
		end		
		if not self.DamageData.Proj then
			self.DamageData.Proj = {}
			self.DamageData.Proj.BP = self:GetBlueprint()
		end
		if self.DamageData.Proj.BP then
			if masterDebug and myDebug then
				LOG('		SetProjBP (/hook/lua/sim/Projectile.lua), Proj BP ', repr( self.DamageData.Proj.BP.BlueprintId ) )
			end		
			return true
		else
			WARN('*** SetProjBP (/hook/lua/sim/Projectile.lua), Proj BP missing or invalid ***')
			return false
		end	
	end,
	
	RtnProjBP = function( self )
		if self.DamageData.Proj.BP then
			return self.DamageData.Proj.BP
		else
			WARN('*** RtnProjBP (/hook/lua/sim/Projectile.lua), Proj BP is missing ***')
			return false 
		end		
	end,	
	
	SetLauncherUnit = function( self )
		local myDebug = false
		if masterDebug and myDebug then
			LOG('	SetLauncherUnit (/hook/lua/sim/Projectile.lua), Projectile ID ', self:GetEntityId() )
		end			
		local launcher = self:GetLauncher()
		self.DamageData.Launcher = {}		
		if IsUnit( launcher ) then
			self.DamageData.Launcher = launcher
		end
		if self.DamageData.Launcher then
			if masterDebug and myDebug then
				LOG('		SetLauncherUnit (/hook/lua/sim/Projectile.lua), projectiles Launcher ID ', self.DamageData.Launcher:GetEntityId() ) 
			end		
			return true
		else
			WARN('*** SetLauncherUnit (/hook/lua/sim/Projectile.lua), Launcher missing or invalid entity ***')
			return false
		end															
	end,
	
	RtnLauncher = function( self)
		if self.DamageData.Launcher then
			return self.DamageData.Launcher
		else
			WARN('*** RtnLauncher (/hook/lua/sim/Projectile.lua), Launcher is missing ***')		
			return false
		end
	end,
	
	SetLauncherBP = function( self )
		local myDebug = false
		if masterDebug and myDebug then
			LOG('	SetLauncherBP (/hook/lua/sim/Projectile.lua), Projectile ID ', self:GetEntityId() )
		end	
		if not self.DamageData.Launcher.BP then
			self.DamageData.Launcher.BP = self:RtnLauncher():GetBlueprint()	
		end	
		if self.DamageData.Launcher.BP then
			if masterDebug and myDebug then
				LOG('		SetLauncherBP (/hook/lua/sim/Projectile.lua), Launcher BP ', repr( self.DamageData.Launcher.BP.BlueprintId or nil ) ) 
			end		
			return true
		else
			WARN('*** SetLauncherBP (/hook/lua/sim/Projectile.lua), Launcher BP missing or invalid ***')	
			return false
		end					
	end,
	
	RtnLauncherBP = function( self )
		if self.DamageData.Launcher.BP then
			return self.DamageData.Launcher.BP
		else
			WARN('*** RtnLauncherBP (/hook/lua/sim/Projectile.lua), Launcher BP is missing ***')
			return false 
		end		
	end,	
	
	SetLauncherWeapon = function( self )
		local myDebug = false
		if masterDebug and myDebug then
			LOG('	SetLauncherWeapon (/hook/lua/sim/Projectile.lua), Projectile ID ', self:GetEntityId() )
		end		
		if not self.DamageData.Weapon then
			self.DamageData.Weapon = {}
			--> Determine what weapon launched this projectile
			if table.getsize( self:RtnLauncherBP().Weapon ) > 0 then
				for k, v in self:RtnLauncherBP().Weapon do
					--> Do the projectile BlueprintId's match ?
					if v.ProjectileId == self:RtnProjBP().BlueprintId then							
						self.DamageData.Weapon = v
						break
					end
				end
				if self.DamageData.Weapon then
					if masterDebug and myDebug then
						LOG('		SetLauncherWeapon (/hook/lua/sim/Projectile.lua), Returning the weapon that fired this projectile ', repr( self.DamageData.Weapon.Label ) )
					end				
					return true
				else
					if self:RtnLauncherBP().Weapon[1] then
						self.DamageData.Weapon = self:RtnLauncherBP().Weapon[1]
						if masterDebug and myDebug then
							LOG('		SetLauncherWeapon (/hook/lua/sim/Projectile.lua), Returning self.DamageData.Weapon[1] as the firing weapon ', repr( self.DamageData.Weapon.Label ) )
						end				
						return true
					else
						WARN('*** SetLauncherWeapon (/hook/lua/sim/Projectile.lua), "self:RtnLauncherBP().Weapon[1]" is nil ***')
						return false					
					end
				end
			else
				WARN('*** SetLauncherWeapon (/hook/lua/sim/Projectile.lua), "self.DamageData.Launcher.BP.Weapon" table size is 0 ***')
				return false
			end
		elseif masterDebug and myDebug then
			LOG('		SetLauncherWeapon (/hook/lua/sim/Projectile.lua), Weapon already set')	
			return true
		end
	end,
	
	RtnLauncherWeapon = function( self )
		if self.DamageData.Weapon then
			return self.DamageData.Weapon
		else
			WARN('*** RtnLauncherWeapon (/hook/lua/sim/Projectile.lua), Weapon is missing ***')
			return false 
		end		
	end,		
	
	SetProjVelocity = function( self )
		local myDebug = false
		if masterDebug and myDebug then
			LOG('	SetProjVelocity (/hook/lua/sim/Projectile.lua), Projectile ID ', self:GetEntityId() )
		end			
		if not self.DamageData.MuzzleVelocity then
			self.DamageData.MuzzleVelocity = {}
			self.DamageData.WeaponCat = {}
			
			--> Determine Projectile Velocity by what "category" and type the firing unit represents
			if self:RtnLauncherWeapon().BallisticArc == 'RULEUBA_HighArc' then
				self.DamageData.WeaponCat = 'Artillery'
			elseif self:RtnLauncherBP().Physics.MotionType == 'RULEUMT_Air' and EntityCategoryContains( categories.BOMBER, self:RtnLauncher() ) then	
				self.DamageData.WeaponCat = 'Bomb'				
			else
				self.DamageData.WeaponCat = 'DirectFire'
			end	
																	
			if self:RtnLauncher() and self:RtnLauncherBP() and self:RtnLauncherWeapon() then
				if self.DamageData.WeaponCat == 'Artillery' then
					local dist, distMulti
					local maxRng = self:RtnLauncherWeapon().MaxRadius
					if self.DamageData.LauncherLoc and maxRng > 0 then
						--> Use the artilleries distance from the target to adjust the projectiles speed										
						dist = utilities.GetDistanceBetweenTwoVectors( self.DamageData.LauncherLoc , self:GetPosition() )
						distMulti = dist / maxRng
					else
						-->Randomly generate a multi if the distance is not known
						distMulti = utilities.GetRandomFloat( 0.25, 0.75 )										
					end		
					if self:RtnLauncherWeapon().Label and self:RtnLauncherWeapon().MuzzleVelocity and ( self:RtnLauncherWeapon().MuzzleVelocity * distMulti ) > minArtyVelocity then				
						--> See if the weapons MuzzleVelocity is valid for use								
						self.DamageData.MuzzleVelocity = self:RtnLauncherWeapon().MuzzleVelocity * distMulti
						if masterDebug and myDebug then
							LOG('	SetProjVelocity (/hook/lua/sim/Projectile.lua), Artillery method 1 ')
						end							 					
					elseif self:RtnLauncherWeapon().Label and self:RtnProjBP().Physics.InitialSpeed and ( self:RtnProjBP().Physics.InitialSpeed * distMulti ) > minArtyVelocity then					
						--> See if the projectiles BP InitialSpeed is valid for use						
						self.DamageData.MuzzleVelocity = self:RtnProjBP().Physics.InitialSpeed * distMulti
						if masterDebug and myDebug then
							LOG('	SetProjVelocity (/hook/lua/sim/Projectile.lua), Artillery method 2 ')
						end							 														
					else					
						--> Should all else fail use the minArtyVelocity					
						self.DamageData.MuzzleVelocity = minArtyVelocity
						if masterDebug and myDebug then
							LOG('	SetProjVelocity (/hook/lua/sim/Projectile.lua), Artillery minArtyVelocity ')
						end						 						
					end							
					if masterDebug and myDebug then
						LOG('		SetProjVelocity (/hook/lua/sim/Projectile.lua), Weapon Label ', self:RtnLauncherWeapon().Label,' ARTILLERY MuzzleVelocity ', self.DamageData.MuzzleVelocity )
					end																				
				elseif self.DamageData.WeaponCat == 'Bomb' then								
					if self:RtnLauncherWeapon().Label and self:RtnLauncherWeapon().MuzzleVelocity and self:RtnLauncherWeapon().MuzzleVelocity > minBombVelocity then
						--> See if the weapons MuzzleVelocity is valid for use					
						self.DamageData.MuzzleVelocity = self:RtnLauncherWeapon().MuzzleVelocity
						if masterDebug and myDebug then
							LOG('	SetProjVelocity (/hook/lua/sim/Projectile.lua), Bomb method 1 ')
						end													 
					elseif self:RtnLauncherWeapon().Label and self:RtnLauncherBP().Air.MaxAirspeed and self:RtnLauncherBP().Air.MaxAirspeed > minBombVelocity then
 						--> See if the bombers MaxAirspeed is valid for use					
						self.DamageData.MuzzleVelocity = self:RtnLauncherBP().Air.MaxAirspeed
						if masterDebug and myDebug then
							LOG('	SetProjVelocity (/hook/lua/sim/Projectile.lua), Bomb method 2 ')
						end							 														
					elseif self:RtnLauncherWeapon().Label and self:RtnProjBP().Physics.MaxSpeed and self:RtnProjBP().Physics.MaxSpeed * 0.5 > minBombVelocity then
						--> See if the bombers projectile BP MaxSpeed is valid for use
						self.DamageData.MuzzleVelocity = self:RtnProjBP().Physics.MaxSpeed * 0.5
						if masterDebug and myDebug then
							LOG('	SetProjVelocity (/hook/lua/sim/Projectile.lua), Bomb method 3 ')
						end														
					else
						--> Should all else fail use the minBombVelocity
						self.DamageData.MuzzleVelocity = minBombVelocity
						if masterDebug and myDebug then
							LOG('	SetProjVelocity (/hook/lua/sim/Projectile.lua), Bomb minBombVelocity ')
						end												 
					end
					if masterDebug and myDebug then
						LOG('		SetProjVelocity (/hook/lua/sim/Projectile.lua), Weapon Label ', self:RtnLauncherWeapon().Label,' BOMBER MuzzleVelocity ', self.DamageData.MuzzleVelocity )
					end				
				else					
					if self:RtnLauncherWeapon().Label and self:RtnLauncherWeapon().MuzzleVelocity and self:RtnLauncherWeapon().MuzzleVelocity > minDirectVelocity then
						--> See if the weapons MuzzleVelocity is valid for use									
						self.DamageData.MuzzleVelocity = self:RtnLauncherWeapon().MuzzleVelocity
						if masterDebug and myDebug then
							LOG('	SetProjVelocity (/hook/lua/sim/Projectile.lua), DirectFire method 1 ')
						end											 					
					elseif self:RtnLauncherWeapon().Label and self:RtnProjBP().Physics.InitialSpeed and self:RtnProjBP().Physics.InitialSpee > minDirectVelocity then
						--> See if the projectiles BP InitialSpeed is valid for use						
						self.DamageData.MuzzleVelocity = self:RtnProjBP().Physics.InitialSpeed
						if masterDebug and myDebug then
							LOG('	SetProjVelocity (/hook/lua/sim/Projectile.lua), DirectFire method 2 ')
						end						 					
					else
						--> Should all else fail use the minDirectVelocity					
						self.DamageData.MuzzleVelocity = minDirectVelocity
						if masterDebug and myDebug then
							LOG('	SetProjVelocity (/hook/lua/sim/Projectile.lua), DirectFire minDirectVelocity ')
						end						 						
					end											
					if masterDebug and myDebug then
						LOG('		SetProjVelocity (/hook/lua/sim/Projectile.lua), Weapon Label ', self:RtnLauncherWeapon().Label,' DIRECTFIRE MuzzleVelocity ', self.DamageData.MuzzleVelocity)
					end
				end				
				--> limit the max speed a projectile can attain
				if self.DamageData.MuzzleVelocity > maxProjVelocity then
					self.DamageData.MuzzleVelocity = maxProjVelocity
				end
				return true
			else
				WARN('*** SetProjVelocity (/hook/lua/sim/Projectile.lua), Unable to detirmine muzzle velocity as either Launcher or Weapon is missing ***')
				return false		
			end
		else
			if masterDebug and myDebug then
				LOG('		SetProjVelocity (/hook/lua/sim/Projectile.lua), Will carry over MuzzleVelocity from current self.DamageData.MuzzleVelocity ', self.DamageData.MuzzleVelocity)
			end
			return true			
		end
	end,
	
	RtnProjVelocity = function( self )
		if self.DamageData.MuzzleVelocity then
			return self.DamageData.MuzzleVelocity
		else
			WARN('*** RtnProjVelocity (/hook/lua/sim/Projectile.lua), Velocity is missing ***')
			return false 
		end		
	end,	

	SetProjLifetime = function( self )
		local myDebug = false
		if masterDebug and myDebug then
			LOG('	SetProjLifetime (/hook/lua/sim/Projectile.lua), Projectile ID ', self:GetEntityId() )
		end			
		if not self.DamageData.ProjLifetime then
			if self:RtnLauncher() and self:RtnLauncherWeapon() and self:RtnProjBP() then						
				if self.DamageData.WeaponCat == 'Artillery' then				
					if self:RtnLauncherWeapon().ProjectileLifetime and self:RtnLauncherWeapon().ProjectileLifetime > minArtyLifeTime then
						--> See if the weapons ProjectileLifetime is valid for use	
						self.DamageData.ProjLifetime = self:RtnLauncherWeapon().ProjectileLifetime 
					elseif self:RtnLauncherWeapon().ProjectileLifetimeUsesMultiplier > 0 and self:RtnProjBP().Physics.Lifetime > 0  and ( self:RtnLauncherWeapon().ProjectileLifetimeUsesMultiplier * self:RtnProjBP().Physics.Lifetime ) > minArtyLifeTime then
						--> See if the weapons ProjectileLifetimeUsesMultiplier and projectiles BP Lifetime is valid for use
						self.DamageData.ProjLifetime = self:RtnLauncherWeapon().ProjectileLifetimeUsesMultiplier * self:RtnProjBP().Physics.Lifetime
					else
						--> Should all else fail use the minArtyLifeTime								
						self.DamageData.ProjLifetime = minArtyLifeTime
					end							
				elseif self.DamageData.WeaponCat == 'Bomb' then
					if self:RtnLauncherWeapon().ProjectileLifetime > minBombLifeTime then
						--> See if the weapons ProjectileLifetime is valid for use					
						self.DamageData.ProjLifetime = self:RtnLauncherWeapon().ProjectileLifetime					
					elseif self:RtnProjBP().Physics.Lifetime and self:RtnProjBP().Physics.Lifetime > minBombLifeTime then
						--> See if the projectiles BP Lifetime is valid for use	
						self.DamageData.ProjLifetime = self:RtnProjBP().Physics.Lifetime
					else
						--> Should all else fail use the minBombLifeTime
						self.DamageData.ProjLifetime = minBombLifeTime
					end													
				elseif self.DamageData.WeaponCat == 'DirectFire' then
					if self:RtnLauncherWeapon().ProjectileLifetime and self:RtnLauncherWeapon().ProjectileLifetime > minDirectLifeTime then
						--> See if the weapons ProjectileLifetime is valid for use					
						self.DamageData.ProjLifetime = self:RtnLauncherWeapon().ProjectileLifetime 
					elseif self:RtnLauncherWeapon().ProjectileLifetimeUsesMultiplier > 0 and self:RtnProjBP().Physics.Lifetime > 0  and ( self:RtnLauncherWeapon().ProjectileLifetimeUsesMultiplier * self:RtnProjBP().Physics.Lifetime ) > minDirectLifeTime then
						--> See if the weapons ProjectileLifetimeUsesMultiplier and projectiles BP Lifetime is valid for use
						self.DamageData.ProjLifetime = self:RtnLauncherWeapon().ProjectileLifetimeUsesMultiplier * self:RtnProjBP().Physics.Lifetime
					else
						--> Should all else fail use the minBombLifeTime				
						self.DamageData.ProjLifetime = minDirectLifeTime
					end					
				end	
				--> limit the max lifetime a projectile can attain
				if self.DamageData.ProjLifetime > maxProjLifeTime then
					self.DamageData.ProjLifetime = maxProjLifeTime
				end				
				if masterDebug and myDebug then
					LOG('		SetProjLifetime (/hook/lua/sim/Projectile.lua), ProjectileLifetime is: ', self.DamageData.ProjLifetime )
				end
				return true	
			else
				WARN('*** SetProjLifetime (/hook/lua/sim/Projectile.lua), Some data is missing, returning false ***')		
				return false
			end
		else
			if masterDebug and myDebug then
				LOG('		SetProjLifetime (/hook/lua/sim/Projectile.lua), Will carry over ProjectileLifetime from DamageData', self.DamageData.ProjLifetime )
			end
			return true		
		end
	end,
		
	RtnProjLifetime = function( self )
		if self.DamageData.ProjLifetime then
			return self.DamageData.ProjLifetime
		else
			WARN('*** RtnProjLifetime (/hook/lua/sim/Projectile.lua), Lifetime is missing ***')	
			return false 
		end		
	end,
	
	SetProjVectors= function( self, targetEntity )
		local myDebug = false
		if masterDebug and myDebug then
			LOG('	SetProjVectors (/hook/lua/sim/Projectile.lua), Projectile ID ', self:GetEntityId() )
		end					
		if self and self:GetVelocity() and targetEntity then
			--> Use the position of the projectile and impacted entity to calculate the correct reflection offset		
			local v1 = self:GetPosition()					
			local v2 = targetEntity:GetPosition()
			local velocity =  Vector( self:GetVelocity() )	
	 		self.DamageData.Proj.ReflectVectors = {}	
			if self.DamageData.WeaponCat == 'Artillery' then
				--> Reflection more dependant on where it impacts the shield, looks better for artillery impacts			
				local dir = utilities.GetDifferenceVector( v1, v2 ) 			
				self.DamageData.Proj.ReflectVectors = utilities.GetDirectionVector( velocity, dir )
			elseif self:RtnLauncherWeapon().UWRC_DirectFire then
				--> Works good for units such as tanks and mechmarines
				self.DamageData.Proj.ReflectVectors = velocity						
			else
				--> Reflection dependant on the angle of impact, looks better for units with short ranged weapons that are not direct fire
				local dir = utilities.GetDirectionVector( v1, v2 )			
				self.DamageData.Proj.ReflectVectors = utilities.GetDifferenceVector( velocity, dir )	
			end 					
			--> Randomize the vectors a bit so the projectile doesnt return down the vector every shot			
 			self.DamageData.Proj.ReflectVectors[1] = utilities.GetRandomFloat( (self.DamageData.Proj.ReflectVectors[1] * 0.950), (self.DamageData.Proj.ReflectVectors[1] * 1.050) )--> Left and Right Spread
 			if self:RtnLauncherWeapon().UWRC_DirectFire then						
 				self.DamageData.Proj.ReflectVectors[2] = utilities.GetRandomFloat( (self.DamageData.Proj.ReflectVectors[2] * 0.250), (self.DamageData.Proj.ReflectVectors[2] * 1.000) )--> Up and Down Spread
 			else
 				self.DamageData.Proj.ReflectVectors[2] = utilities.GetRandomFloat( (self.DamageData.Proj.ReflectVectors[2] * 0.750), (self.DamageData.Proj.ReflectVectors[2] * 1.250) )--> Up and Down Spread
 			end	
			if masterDebug and myDebug then
				LOG('		SetProjVectors (/hook/lua/sim/Projectile.lua), Projectile vectors ', repr(self.DamageData.Proj.ReflectVectors) )
			end						
			--> Returning "true" bypasses shield impact and allows reflection
			return true
		else
			WARN('*** SetProjVectors (/hook/lua/sim/Projectile.lua), Some data is missing, Returning false ***')
			--> Returning "false" allows shield impact and bypasses reflection
			return false
		end
	end,
	
	RtnProjVectors = function( self )
		if self.DamageData.Proj.ReflectVectors then
			return self.DamageData.Proj.ReflectVectors
		else
			WARN('*** RtnProjVectors (/hook/lua/sim/Projectile.lua), Vectors are missing ***')
			return false 
		end		
	end,
	
	SetReflected = function( self, entity )
		local myDebug = false
		if not self.DamageData.LastReflectorImpacted and entity then
			self.DamageData.LastReflectorImpacted = {}
			--> Set the impacted shield as the reflector entity
			self.DamageData.LastReflectorImpacted = entity
			if masterDebug and myDebug then
				LOG('	SetReflected (/hook/lua/sim/Projectile.lua), Projectile reflected from ', self.DamageData.LastReflectorImpacted )
			end			
		else
			if masterDebug and myDebug then
				LOG('	SetReflected (/hook/lua/sim/Projectile.lua), SetReflected was previously set')
			end			
		end
	end,
	
	RtnReflected = function( self )
		local myDebug = false
		--> Return what the projectile last impacted
		if self.DamageData.LastReflectorImpacted then
			if masterDebug and myDebug then
				LOG('	RtnReflected (/hook/lua/sim/Projectile.lua), Projectile has reflected once before, returning the entity')
			end
			return self.DamageData.LastReflectorImpacted
		else
			if masterDebug and myDebug then
				LOG('	RtnReflected (/hook/lua/sim/Projectile.lua), Projectile has not reflected yet, returning false')
			end
			return false
		end		
	end,
			
	ProjReflect = function( self, targetEntity )
		local myDebug = false
		if masterDebug and myDebug then
			LOG('*** ProjReflect (/hook/lua/sim/Projectile.lua), Projectile ID ', self:GetEntityId(),' ***')
		end
		--> Allow the shields owner to get credit for the damage and thus gain veterancy
		self:SetInstigator( targetEntity )		
		--> Allow the projectile to damage what originally fired it
		self.DamageData.CollideFriendly = true
		self.DamageData.DamageFriendly = true
		self.DamageData.DamageSelf = true
		--> Lower the base damage of the projectile to 50% of original
		self.DamageData.DamageAmount = self.DamageData.DamageAmount * 0.5
		--> Ensure the damage can never go lower than minDamage
		if self.DamageData.DamageAmount < minDamage then
			self.DamageData.DamageAmount = minDamage
		end			
		--> Reflect the projectile
		self:SetVelocityAlign( true )
		self:TrackTarget( false )
		self:SetVelocity( -self:RtnProjVectors()[1], -self:RtnProjVectors()[2], -self:RtnProjVectors()[3] )
		self:SetVelocity( self:RtnProjVelocity() or minProjVelocity )
		self:SetLifetime( self:RtnProjLifetime() or minProjLifeTime )
		--> Remove projectiles that have reached the end of their lifetime
		if AllowProjTimeOut then
			if not self.ProjTimeoutThread == nil then
				KillThread( self.ProjTimeoutThread )
				self.ProjTimeoutThread = nil
			end
			self:ForkThread( self.ProjTimeoutThread, self )
		end		
		--> Set the original launcher as the new target for the reflected projectile
		if self:RtnLauncher() then
			self:SetNewTarget( self:RtnLauncher() )
		end
	end,
	
	SetInstigator = function( self, entity )
		local myDebug = false
		if entity then
			--> Set the impacted shield generator as the damaging entity for the projectile
			self.DamageData.NewInstigator = entity.Owner or self		
			if masterDebug and myDebug then
				LOG('	SetInstigator (/hook/lua/sim/Projectile.lua), New instigator set for Projectile ID ', self:GetEntityId() )
			end
		else
			WARN('*** SetInstigator (/hook/lua/sim/Projectile.lua), entity failed to pass or is missing ***')
		end
	end,
	
	RtnInstigator = function( self )
		local myDebug = false
		--> Return what the projectile last impacted
		if self.DamageData.NewInstigator then
			if masterDebug and myDebug then
				LOG('	RtnInstigator (/hook/lua/sim/Projectile.lua), retuning New Instigator')
			end
			return self.DamageData.NewInstigator
		else
			WARN('*** RtnInstigator (/hook/lua/sim/Projectile.lua), Instigator not avalible, returning false***')
			return false
		end	
	end,
	
    ProjTimeoutThread = function( self )
		local myDebug = false
		if masterDebug and myDebug then
			LOG('	ProjTimeoutThread (/hook/lua/sim/Projectile.lua) ', self:RtnProjLifetime() )
		end    
        WaitSeconds( self:RtnProjLifetime() or 1 )
        self:Destroy()
    end,	
		
	OnImpact = function( self, targetType, targetEntity )
		local myDebug = false
		if targetEntity then
			--> Prevent multible impacts from the same projectile
			if self.LastImpact == targetEntity then
				if masterDebug and myDebug then
					LOG('	OnImpact (/hook/lua/sim/Projectile.lua), Bypassing oldProjectile via Destroy()')
				end	
				--> Bypass oldProjectile and destroy						
				self:Destroy()				
			else
				if masterDebug and myDebug then
					LOG('	OnImpact (/hook/lua/sim/Projectile.lua), LastImpact set')
				end				
				self.LastImpact = {}																
				self.LastImpact = targetEntity
			end
		end
		if self.DamageData.DamageAmount > 0 then
			if targetType == 'Shield' and targetEntity == self:RtnReflected() then
				if masterDebug and myDebug then
					LOG('		OnImpact (/hook/lua/sim/Projectile.lua), Running shield damage and effects')
				end			
				--> Do damage to the shield
				self:DoDamage( self:RtnLauncher(), self.DamageData, targetEntity )
				self:DoMetaImpact( self.DamageData )
				self:DoUnitImpactBuffs( targetEntity )
			else
				--> Run oldProjectile
				if masterDebug and myDebug then
					LOG('		OnImpact (/hook/lua/sim/Projectile.lua), Running oldProjectile')
				end						
				oldProjectile.OnImpact( self, targetType, targetEntity )		
			end
		else
			--> Destroy projectiles that do no damage 
			self:Destroy()
		end
	end,

	DoDamage = function( self, instigator, damageData, targetEntity )
		local myDebug = false
		--> Run oldProjectile
		if masterDebug and myDebug then
			LOG('	DoDamage (/hook/lua/sim/Projectile.lua), Running oldProjectile.DoDamage')
		end
		if damageData.DamageAmount > minDamage and self:RtnReflected() and self:RtnInstigator() then
			instigator = self:RtnInstigator()
		end
		--> Run oldProjectile		
		oldProjectile.DoDamage( self, instigator, damageData, targetEntity )
	end,

	CreateImpactEffects = function( self, army, EffectTable, EffectScale )
		local myDebug = false
		if masterDebug and myDebug then
			LOG('	CreateImpactEffects (/hook/lua/sim/Projectile.lua), Running oldProjectile.CreateImpactEffects')
		end
		--> Run oldProjectile			
		oldProjectile.CreateImpactEffects( self, army, EffectTable, EffectScale )
	end,

	CreateTerrainEffects = function( self, army, EffectTable, EffectScale )
		local myDebug = false
		--> Run oldProjectile
		if masterDebug and myDebug then
			LOG('	CreateTerrainEffects (/hook/lua/sim/Projectile.lua), Running oldProjectile.CreateTerrainEffects')
		end
		oldProjectile.CreateTerrainEffects( self, army, EffectTable, EffectScale )
	end,

	OnLostTarget = function( self )
		local myDebug = false
		if masterDebug and myDebug then
			LOG('	OnLostTarget (/hook/lua/sim/Projectile.lua), Projectile ID ', self:GetEntityId() or nil )
		end
		--> Run oldProjectile
		oldProjectile.OnLostTarget( self )
	end,

	OnKilled = function( self, instigator, type, overkillRatio )
		local myDebug = false
		if masterDebug and myDebug then
			LOG('	OnKilled (/hook/lua/sim/Projectile.lua), Projectile ID ', self:GetEntityId() or nil )
		end
		--> Run oldProjectile
		oldProjectile.OnKilled( self, instigator, type, overkillRatio )
	end,

	OnDestroy = function(self)
		local myDebug = false
		if masterDebug and myDebug then
			LOG('OnDestroy (/hook/lua/sim/Projectile.lua), Projectile ID ', self:GetEntityId() or nil )
		end
		--> Run oldProjectile
		oldProjectile.OnDestroy( self )
	end,

}
end--(end of non-destructive hook)